using System;
using System.CodeDom;
using System.CodeDom.Compiler;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using Microsoft.CSharp;

namespace TestCalc {
	public class Evaluator {
		#region Variables
		public List<String> ErrorList = new List<String>();
		private MethodInfo _mi;
		private Object _assemblyInstance;
		#endregion

		#region Prepare, Evaluate
		public Boolean Prepare(String formula) {
			// Sample of formula
			//String formula = "curve[0] + curve[1] * curve[2] / curve[3] - curve[4]";
			String expression = this._preprocessFormula(formula);
			String code = this._buildClass(expression);

			CodeDomProvider codeProvider = null;
			codeProvider = new CSharpCodeProvider();
			ICodeCompiler compiler = codeProvider.CreateCompiler();
			// Get compiler parameters
			CompilerParameters compilerParams = new CompilerParameters();
			compilerParams.CompilerOptions = "/target:library /optimize";
			compilerParams.GenerateExecutable = false;
			compilerParams.GenerateInMemory = true;
			compilerParams.IncludeDebugInformation = false;
			compilerParams.ReferencedAssemblies.Add("mscorlib.dll");
			compilerParams.ReferencedAssemblies.Add("System.dll");
			CompilerResults results = compiler.CompileAssemblyFromSource(compilerParams, code);
			// Exit on errors
			if (results.Errors.Count > 0) {
				foreach (CompilerError error in results.Errors) { this.ErrorList.Add(error.ErrorText); }
				return false;
			}

			this._assemblyInstance = results.CompiledAssembly.CreateInstance("FormulaEvaluator.Calculator");
			this._mi = results.CompiledAssembly.GetModules(false)[0].GetTypes()[0].GetMethods()[0];
				
			return true;
		}

		public Double Evaluate(Double[] curve) {
			return (Double) this._mi.Invoke(this._assemblyInstance, new Object[] { curve });
		}
		#endregion

		#region Preprocessor
		// Need to change formula to use .NET Math library
		private String _preprocessFormula(String formula) {
			List<String> mathFunctions = new List<String>();
			Dictionary<String, String> mathFunctionsMap = new Dictionary<String, String>();
			Assembly systemAssembly = Assembly.GetAssembly(typeof(System.Math));
			if (systemAssembly != null) {
				Module[] modules = systemAssembly.GetModules(false);
				Type[] types = modules[0].GetTypes();
				foreach (Type type in types) {
					if (type.Name != "Math") { continue; }
					MemberInfo[] mis = type.GetMembers();
					foreach (MemberInfo mi in mis) {
						if (mi.Name.ToUpper() != "E") {
							mathFunctions.Add(mi.Name);
							mathFunctionsMap[mi.Name.ToUpper()] = "Math." + mi.Name;
						}
					}
				}
			}
			mathFunctions.Add("Int");
			mathFunctionsMap["Int".ToUpper()] = "Math.Floor";
			mathFunctions.Add("Mod");
			mathFunctionsMap["Mod".ToUpper()] = "%";

			Regex re = new Regex("[a-zA-Z_]+");
			ArrayList replacelist = new ArrayList();
			MatchCollection matches = re.Matches(formula);
			foreach (Match m in matches) {
				if (!replacelist.Contains(m.Value) && mathFunctionsMap.ContainsKey(m.Value.ToUpper())) {
					formula = formula.Replace(m.Value, mathFunctionsMap[m.Value.ToUpper()]);
				}
				replacelist.Add(m.Value);
			}

			// return the modified evaluation String
			return formula;
		}
		
		// Build code dom
		private String _buildClass(String formula) {
			// need a String to put the code into
			StringBuilder source = new StringBuilder();
			StringWriter sw = new StringWriter(source);

			// Declare your provider and generator
			CSharpCodeProvider codeProvider = new CSharpCodeProvider();
			ICodeGenerator codeGenerator = codeProvider.CreateGenerator(sw);
			CodeGeneratorOptions codeGenOpts = new CodeGeneratorOptions();

			CodeNamespace myNS = new CodeNamespace("FormulaEvaluator");
			myNS.Imports.Add(new CodeNamespaceImport("System"));

			// Build the class declaration and member variables			
			CodeTypeDeclaration classDeclaration = new CodeTypeDeclaration();
			classDeclaration.IsClass = true;
			classDeclaration.Name = "Calculator";
			classDeclaration.Attributes = MemberAttributes.Public;

			// Our calculate method
			CodeMemberMethod myMethod = new CodeMemberMethod();
			myMethod.Name = "Calculate";
			myMethod.ReturnType = new CodeTypeReference(typeof(Double));
			myMethod.Attributes = MemberAttributes.Public | MemberAttributes.Final;
			myMethod.Parameters.Add(new CodeParameterDeclarationExpression(new CodeTypeReference(typeof(Double[])), "curve"));
			myMethod.Statements.Add(new CodeMethodReturnStatement(new CodeSnippetExpression(formula)));
			classDeclaration.Members.Add(myMethod);
			// Write code
			myNS.Types.Add(classDeclaration);
			codeGenerator.GenerateCodeFromNamespace(myNS, sw, codeGenOpts);
			sw.Flush();
			sw.Close();
			return source.ToString();
		}
		#endregion

	}
}
